require 'rest-client'
require 'json'

module CaptchaSolver
  class Base
    class NotStartedError < StandardError; end
    class FailedError < StandardError; end
    class NotSolvedError < StandardError; end

    TIMEOUTS_BETWEEN_ATTEMPTS = [40, 40, 40, 40, 40]

    attr_reader :rucaptcha_api_key, :proxy_address,
      :start_response, :complete_response

    def initialize(args = {})
      @rucaptcha_api_key = ENV['rucaptcha_api_key']
      @proxy_address = args[:proxy_address]
    end

    # 1. Makes a in.php request in order to get the rucaptcha id.
    # 2. Waits for 40 seconds, so the captcha can be solved by that time.
    # 3. Makes a res.php request in order to get the solved captcha.
    # 4. If captcha is not solved, it tries to do the same 4 more times.
    def solve
      start
      raise NotStartedError.new(start_response&.body) unless started?

      TIMEOUTS_BETWEEN_ATTEMPTS.each do |seconds|
        wait_for(seconds)

        complete

        raise FailedError.new(complete_response&.body) if failed?
        return solved_captcha if solved?
      end

      raise NotSolvedError.new(complete_response&.body)
    end

    # Returns solved captcha.
    def solved_captcha
      return nil unless solved?
      captcha_value
    end

    private

    # Headers that have to be passed in each request.
    def default_request_headers
      { 'Accept' => 'application/json', 'Content-Type' => 'application/x-www-form-urlencoded' }
    end

    # Parameters that have to be passed in each request.
    def default_request_params
      { key: rucaptcha_api_key, json: 1 }
    end

    # Parameters which are used in in.php request.
    def start_request_params
      raise NotImplementedError.new('#start_request_params method must be implemented')
    end

    # Makes a in.php request and saves the result in @start_response.
    def start
      @start_response = RestClient::Request.execute(
        method: :post,
        url: 'http://rucaptcha.com/in.php',
        proxy: "http://#{proxy_address}",
        headers: default_request_headers,
        payload: default_request_params.merge(start_request_params)
      )
    end

    # Parses response from in.php request and returns rucaptcha id.
    def rucaptcha_id
      return nil if start_response.nil?
      JSON.parse(start_response.body)['request']
    end

    # Checks if a in.php request is successful.
    def started?
      return false if start_response.nil?
      return false if errors_regexp.match?(start_response.body)
      !rucaptcha_id.nil?
    end

    # Parameters which are used in res.php request.
    def complete_request_params
      { id: rucaptcha_id, action: 'get' }
    end

    # Makes a res.php request and saves the result in @complete_response.
    def complete
      return nil unless started?
      @complete_response = RestClient::Request.execute(
        method: :post,
        url: 'http://rucaptcha.com/res.php',
        proxy: "http://#{proxy_address}",
        headers: default_request_headers,
        payload: default_request_params.merge(complete_request_params)
      )
    end

    # Checks if captcha solving is failed.
    def failed?
      return false if complete_response.nil?
      errors_regexp.match? complete_response.body
    end

    # Checks if captcha is solved.
    def solved?
      return false if complete_response.nil?
      return false if failed?
      !not_ready_regexp.match?(complete_response.body) && !captcha_value.nil?
    end

    # Parses response from res.php request and returns captcha value.
    def captcha_value
      JSON.parse(complete_response.body)['request']
    end

    # The error that is returned by res.php request if captcha is not solved yet.
    def not_ready_regexp
      /CAPCHA_NOT_READY/i
    end

    # Any error that might be returned by in.php or res.php request.
    def errors_regexp
      /(ERROR|BANNED|MAX_USER_TURN|REPORT_NOT_RECORDED)/i
    end

    # Makes a delay between requests.
    def wait_for(seconds)
      puts "Waiting for #{seconds} seconds"
      seconds.times do |i|
        sleep(1)
        print '.'
      end
      puts
    end
  end
end
